/* N-body calculation using Greenspan 1st-order discrete mechanics.	*/
/* Uses N*N*Q processors, communication by spreading and reduction,	*/
/* explict loops for xyz.  All processors store particle information	*/
/* and calculate velocities, positions.					*/
/* Uses polystep technique, integrating Q time steps in parallel.	*/
/* Uses dynamic shape allocation to permit more flexible use.		*/
/* Uses some paris code for spreading.					*/
/* File:  nb1dpoly.cs							*/

#include <stdio.h>
#include <math.h>
#include <float.h>
#include <stdlib.h>
#include <cscomm.h>
#include <cm/timers.h>

/* Max number of bodies in the calculation.  Q = 4, N = 64 => 16k procs */
/* Q, N must be powers of 2, Q*N^2 >= physical CM size.			*/

int Q  =  4;	/* Number of parallel time steps.		*/
int Q0 =  0;	/* Index of block with current time step	*/
int N  = 64;	/* Max number of bodies permitted		*/
int   NB;	/* Number of bodies in system, <= N		*/
		/*****   NB = N during integration	    *****/
int   PX;	/* Controls how much state is printed		*/
int PD = 0;	/* Controls diagnostic printing (0 = off)	*/
float G;	/* Gravitational constant			*/
int advance;	/* Indicates that time step should be advanced.		*/

#define TWOPI      6.2831853f

/* Particles, and pairs of particles.  All entries have particle	*/
/* states.  Some variables could be overloaded. 			*/

shape [][][]pair;	/* Will be [Q][N][N]	*/

/* All parallel variables are declared as pointers so 'pair' can be	*/
/* dynamic.  State variables will be allocated as x0[3], etc.;  w, m,	*/
/* mj, r0, are floats							*/

/* x0, x1 - current, new positions;	v0, v1 - current, new velocities */
/* m - mass;		mj - scaled mass;	ac - acceleration	 */
/* w, ax - working	r0 - distance (X0)	d0 - displacement (X0)	 */

float:pair *x0, *x1, *v0, *v1, *m, *mj, *ac, *d0, *r0, *ax, *w;

/* i - row number;    j - column number;    c - coordinate number	*/

int:pair *i, *j, *c;

/* sm   - processor is part of computation = diag | act	*/
/* diag - processor is on diagonal			*/
/* act  - processor is active pair (off-diag)		*/
/* sp   - block calculating current time step		*/
/* bp   - blocks that are not the most advanced	in time	*/

bool:pair *sm, *diag, *act, *sp, *bp;

/* Allocate global parallel variables, pair must have been allocated.	*/
/* Sizes used elsewhere are global.					*/

int vsz;	/* Size of float 3-vector */

void create_globals()
 { int fsz, isz, bsz;
   fsz = boolsizeof(float:pair);  vsz = 3*fsz;
   isz = boolsizeof(int:pair);
   bsz = boolsizeof(bool:pair);
   /* Float vectors[3] */
   x0 = palloc(pair, vsz);  x1 = palloc(pair, vsz);
   v0 = palloc(pair, vsz);  v1 = palloc(pair, vsz);
   ac = palloc(pair, vsz);  ax = palloc(pair, vsz);
   d0 = palloc(pair, vsz);
   /* Floats */
   m  = palloc(pair, fsz);  mj = palloc(pair, fsz);
   r0 = palloc(pair, fsz);  w  = palloc(pair, fsz);
   /* Ints and bools */
   i  = palloc(pair, isz);  j = palloc(pair, isz);  c = palloc(pair, isz);
   sm  = palloc(pair, bsz);  diag = palloc(pair, bsz);
   act = palloc(pair, bsz);  sp   = palloc(pair, bsz);
   bp  = palloc(pair, bsz);
 }

/* Pairs are represented symmetrically.	 Setup initializes these.	*/

void setup()
 { with (pair)
    { /* Initialize coordinates */
      *c = pcoord(0);  *i = pcoord(1);  *j = pcoord(2);
      /* Identify active processors */
      *sm = (*i < NB) & (*j < NB);
      *act = *sm & (*i != *j);  *diag = *sm & (*i == *j);
    }
 }

FILE *datap;      /* Stream for reading data */

/* Load inital positions, velocities, also masses, gravitational const. */
/* Leaves data for P[i] in processors [0][i][*].			*/

void init()
 { int k;  float ax, ay, az;
   Q0 = 0;	/* Block 0 will be initial time step */
   advance = 0;	/* We aren't ready to advance time step */
   with (pair)
   {*sp = (*c == Q0);  *bp = (*c != (Q - 1));
    for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &ax, &ay, &az);
       [0][k][0]x0[0] = ax;  [0][k][0]x0[1] = ay;  [0][k][0]x0[2] = az;
     };
    for ( k = 0; k < NB; k++ )
     { fscanf(datap, "%e%e%e", &ax, &ay, &az);
       [0][k][0]v0[0] = ax;  [0][k][0]v0[1] = ay;  [0][k][0]v0[2] = az;
     };
    for ( k = 0; k < NB; k++ )
      { fscanf(datap, "%e", &ax);  [0][k][0]*m = ax; };
    fscanf(datap, "%e", &G);
    /* Spread data across rows */
    x0[0] = copy_spread(&x0[0], 2, 0);
    x0[1] = copy_spread(&x0[1], 2, 0);
    x0[2] = copy_spread(&x0[2], 2, 0);
    v0[0] = copy_spread(&v0[0], 2, 0);
    v0[1] = copy_spread(&v0[1], 2, 0);
    v0[2] = copy_spread(&v0[2], 2, 0);
    *m  = copy_spread(m,  2, 0);
   }
 }

/* Insert nx asteroids at radius ar with total mass am.  This assumes	*/
/* that an orbit of radius 1 about origin corresponds to velocity 2*pi	*/
/* (e.g., au-earth-yr units).  Uses parallel trig.			*/

void mkast(int nx, float ar, float am)
 { float:pair sx, cx, arg;
   float spd;  int nn;
   nn = NB + nx;
   spd = TWOPI/sqrt(ar);
   printf("Making %d asteroids, r = %9.5f, spd = %9.5f, tot mass = %9.2e\n",
		  nx,		     ar,	  spd,		    am);
   with (pair) 
   { *sm = (*i < nn) & (*j < nn);
     *act = *sm & (*i != *j);  *diag = *sm & (*i == *j);
     where ( (NB <= *i) & (*i < nn) & *sm )
      { arg = (TWOPI*(*i - NB))/nx;
        sx = sin(arg);  cx = cos(arg);
        x0[0] =  cx*ar;  v0[0] = (-sx)*spd;
        x0[1] =  sx*ar;  v0[1] =  cx*spd;
        x0[2] = 0.0;     v0[2] = 0.0;
        *m = am/nx;
      }}  /* end where, end with (pair) */
   NB = nn;
 }


/* Print current state of system according to PX.  See comments for	*/
/* main program for PX settings.  Takes `advance' into account in	*/
/* determining which values to print.					*/

void print_state()
 { int k, np;  float:current *x, *v;
   if ( (PX >= 0) && (PX < NB) ) np = PX; else np = NB;
   if ( advance ) { x = x1;  v = v1; } else { x = x0;  v = v0; }
   printf(
  "\n       x         y         z         vx        vy        vz        m\n");
   for ( k = 0; k < np; k++ )
     printf("%2d %9.5f %9.5f %9.5f %9.5f %9.5f %9.5f %9.2e\n",
            k, [Q0][k][k]x[0], [Q0][k][k]x[1], [Q0][k][k]x[2],
               [Q0][k][k]v[0], [Q0][k][k]v[1], [Q0][k][k]v[2],
	       [Q0][k][k]*m					 );
 }

/* Print x-components of xx in nx by nx processors (for testing). */
/* Values printed from time slice q >= 0 (typically Q0)		  */
/* nx = 7 fits nicely on one line.				  */
/******************************************************************
void print_x(char lbl[], float:current *xx, int nx, int q)
 { int u, v;
   if ( q >= Q ) q = q%%Q;	*//* Keep in range *//*
   if (PD)
    { printf("\n%s  -  slice %d\n", lbl, q); 
      for ( u = 0; u < nx; u++)
       { for ( v = 0;  v < nx; v++)
           printf("%9.2e ", [q][u][v]*xx);
         printf("\n");
       }
    }
 }
*******************************************************************/
/* Print x-components of bb in nx by nx processors (for testing). */
/* Values printed from time slice q >= 0 (typically Q0)		  */
/******************************************************************
void print_b(char lbl[], bool:current *xx, int nx, int q)
 { int u, v;
   if ( q >= Q ) q = q%%Q;	*//* Keep in range *//*
   if (PD)
    { printf("\n%s  -  slice %d\n", lbl, q); 
      for ( u = 0; u < nx; u++)
       { for ( v = 0;  v < nx; v++)
           printf("%1d ", [q][u][v]*xx);
         printf("\n");
       }
    }
 }
*******************************************************************/

/* Spread xv[3] along i-axis (1-axis) from diagonal, leaving result in dv, */
/* Called with all processors active and leaves all active.  Uses ax.	   */

void spreadi(float:pair *dv, float:pair *xv)
 { int u;
   /* if (PD) printf("In spreadi\n");			For testing	*/
   for ( u = 0;  u < 3;  u++ )
    { where (*diag) ax[u] = xv[u]; else ax[u] = -FLT_MAX;
      dv[u] = spread(ax[u], 1, CMC_combiner_max);
    }
   /* if (PD) printf("Leaving spreadi\n");		For testing	*/
 }

/* Calculate accelerations.  r0, d0 already give displacement, distance  */
/* for current position.  Called with all processors active.		 */
/* 'vsz' is size of float 3-vector, defined above.			 */

void accel()
 { float:pair d1[3], r1;  int u;
 everywhere
 { *w = 0.0f;			/* To accumulate sum of squares (later) */
   for ( u = 0;  u < 3;  u++ ) ac[u] = 0.0f;
   spreadi(&d1[0], x1);		/* Spread x1 to d1 */
   /* print_x("accel, d1 after spread 1", &d1[0], 7, Q0);	For testing */
   /* Now d1 = [c][j][j]x1,  x1 = [c][i][i]x1 */
   where (*act) 
    { for ( u = 0;  u < 3;  u++ )
       { r1 = d1[u] -= x1[u];		/* New displacement component */
         *w += r1*r1;			/* For sum of squares */
       }
      r1 = sqrt(*w);		/* Euclidean distance */
    } /* end where (*act) ... */
   /* print_x("accel, r1", &r1, 7, Q0);	For testing */
   if (advance)		/* Advance time step	*/
    { Q0++;  if ( Q0 == Q ) Q0 = 0;  advance = 0;
      to_torus_dim(x0, x1, vsz, 0, 1);  to_torus_dim(v0, v1, vsz, 0, 1);
      to_torus_dim(d0, d1, vsz, 0, 1);  to_torus_dim(r0, r1, 0, 1);
      where (*sp)	/* Block which just finished time step */
       { for ( u = 0;  u < 3;  u++ )
	  { x1[u] = x0[u];  v1[u] = v0[u];  d1[u] = d0[u]; }
	 r1 = *r0;
       }
      *sp = (*c == Q0);  *bp = (*c != (Q0+Q-1)%Q);	/* Advance sp, bp */
    }	/* End "if advance" case */
   else	/* Propagate x1, etc. */
    { where ( *bp )
       { to_torus_dim(x0, x1, vsz, 0, 1);  to_torus_dim(v0, v1, vsz, 0, 1);
	 to_torus_dim(d0, d1, vsz, 0, 1);  to_torus_dim(r0, r1, 0, 1);
       }
    }	/* End "else" case */
   where (*act) {  *w = *mj/(*r0*r1*(*r0+r1)); }   /* Scalar multiplier */
   for ( u = 0;  u < 3;  u++ )
     { ac[u] = *w*(d0[u]+d1[u]);	/* Acceleration component */
       ac[u] = spread(ac[u], 2, CMC_combiner_add);  /* Sum accelerations */
     }
   /* print_x("accel, final ac", ac, 7, Q0);	For testing */
  } /* end everywhere */
  }

/* Greenspan's 1st-order method.  Performs ns steps of time dt		*/
/* with solution tolerance eps.  Returns number of evaluations.		*/
/* Leaves number of iterations which were stopped at limit in cfail.	*/
/* All time steps should be initialized with appropriate values for	*/
/*        x0, v0, d0, r0, x1, v1					*/
/* '0' items should be spread over all processors in current time step;	*/
/* sp marks the block for the current time step, in which "0" items are	*/
/* fixed.  Some diagnostic printing is provided in comments.		*/


int cfail;   /* Counts number of iterations that failed to converge */

int green1(int ns, float dt, float eps)
 { int k, ic, v, lim=10, u;  float hdt, mx;  float:pair nx, px;
   v = 0;  hdt = 0.5*dt;
   with (pair) everywhere
   { /* print_x("green1, x0 initially", x0, 7, Q0);	For testing */
     /* print_x("green1, v0 initially", v0, 7, Q0);	For testing */
     for (k = 0; k < ns; k++)
       { ic = 0;
	 /* This is a do - while - repeat loop: */
   	 do
          { accel();  v++;  ic++;
	    px = 0.0f;	/* For max change */
	    where (*sm)
	    { for ( u = 0;  u < 3;  u++ )
	       { nx = v0[u] + dt*ac[u];			/* New velocity	*/
		 px >?= fabs(nx - v1[u]);		/* Max change	*/
		 v1[u] = nx;
		 nx = x0[u] + hdt*(v0[u] + v1[u]);	/* New position	*/
		 px >?= fabs(nx - x1[u]);		/* Max change	*/
		 x1[u] = nx; 
	       } /* end for u */
	      where (*sp) mx = (>?= px);   /* Global max in current step*/
  	      /* print_x("green1, v1", v1, 7, Q0);	For testing */
   	      /* print_x("green1, x1", x1, 7, Q0);	For testing */
	      } /* end where (*sm) */
	    if ( (mx < eps) || ( ic == lim ) ) break;
	   }
	 while (1);
         advance = 1;	/* Will advance time step in next accel */
	 if (ic == lim) cfail++;
       } ;   /* end for */
    }  /* end  with (pair) */
   return(v);
 }

/* Initialize masses and system state for green1.	*/

void initgreen()
 { int u;  float:pair r1;
   with (pair) everywhere
    { /* Spread state to forward blocks */
      for ( u = 0;  u < 3;  u++ )
       { x0[u] = copy_spread(&x0[u], 0, Q0);
	 v0[u] = copy_spread(&v0[u], 0, Q0);
       }
      *m = copy_spread(m, 0, Q0);
      /* Spread masses for particle j and scale */
      *m = [*c][*j][*j]*m;  *mj = G**m;
      spreadi(d0, x0);	/* For initial displacements */
      /* print_x("initgreen, d0 after spread", d0, 7);	For testing */
      *w = 0.0f; 	/* For initial distances */
      for ( u = 0;  u < 3;  u++ )
       { r1 = d0[u] -= x0[u];  *w += r1*r1; }
      *r0 = sqrt(*w);
      /* print_x("initgreen, d0 with displacements", d0, 7); For testing */
      /* print_x("initgreen, r0", r0, 7);		For testing */
      for ( u = 0;  u < 3;  u++ )
       { x1[u] = x0[u];  v1[u] = v0[u]; }
    }  /* End with (pair) */
 }

/* Main program:  reads problem and control information from data file	*/
/* named by user and performs calculation.				*/
/* Print control:							*/
/*	  0 - just performance data					*/
/*	< 0 - all							*/
/*	> 0 - number given						*/

void main()
 { int ns, np, k, vl;  float dt, eps, ar, am;
   double eltm, cmtm;		/* Elapsed, CM times */
   char fnm[50] , dgp[50] ;	/* dgp for testing */
   /* Initialize the pair shape with suitable geometry */
   printf("Enter Q, N for [Q][N][N] shape> ");
   scanf("%d%d", &Q, &N);
   allocate_shape(&pair, 3, Q, N, N);
   create_globals();
   printf("Shape is [%d][%d][%d] = %d vp / %d physical, vsz = %d\n",
                     Q,  N,  N, Q*N*N, positionsof(physical), vsz );
   /* Loop for test cases */
   while ( printf("Data file (or q)> "),
	   scanf("%s", fnm),
	   ( fnm[0] != 'q' ) || ( fnm[1] > 0 ) )
    { printf("Print ctl> ");  scanf("%d", &PX);
      printf("Diagnostic output? (y or n)> ");	/* For testing */
      scanf("%s", dgp);				/* For testing */
      if ( dgp[0] == 'y' ) PD = 1; else PD = 0;
      /* Performs np integrations, each ns steps of dt */
      datap = fopen(fnm, "r");
      fscanf(datap, "%d", &NB);
      setup();   /* Setup geometric structure */
      printf("Initialize....\n");
      init();    /* Read data for individual bodies */
      /* Read "asteroid" descriptions */
      while ( fscanf(datap, "%d", &ns), ns > 0)
       { fscanf(datap, "%e%e", &ar, &am);  mkast(ns, ar, am); }
      if ( NB < N ) mkast(N-NB, 3.113, 0.01);	/* Fill with extra asteroids */
      fscanf(datap, "%d%d%e%e", &ns, &np, &dt, &eps);
      fclose(datap);
      printf("\n   NB = %d", NB);
      printf("\n   G = %13.6e  delta-t = %12.5e  epsilon = %12.5e\n",
                    G, dt, eps);
      printf("\n simulated time = 0");
      print_state();
      initgreen();	/* Set up for integration */
      for ( k = 0; k < np; k++ )
       { CM_timer_start(1);
         cfail = 0;
         vl = green1(ns, dt, eps);
         CM_timer_stop(1);
	 eltm = CM_timer_read_elapsed(1);  cmtm = CM_timer_read_cm_busy(1);
	 printf(
	  "\nElapsed: %6.3f sec   CM: %6.3f sec   %3.0f%% utilization\n",
		       eltm,           cmtm,  100.0*(cmtm/eltm) );
	 eltm = 1000*eltm;  cmtm = 1000*cmtm;	/* Now in msec. */
	 printf("%3.1f : %3.1f msec/eval,  %3.1f : %3.1f msec/step\n",
		eltm/vl, cmtm/vl,	  eltm/ns, cmtm/ns );
	 CM_timer_clear(1);
         printf(
        "\n simulated time = %9.3f  %d evaluations  %d iterations limited\n",
	             (k + 1)*ns*dt, vl,            cfail);
         print_state();
       };  /* end for */
    }  /* end while */
 }

/*********************************************************************
  Data files:  as for nb3dg2.cs
*************************************************************************/
